"""This module defines several policies and worlds for use in testing.

Policies:
    NOOP_POLICY: A policy that represents no operation, with all zeros.
    FORWARD_POLICY: A policy that represents moving forward, with a value of 1 in the first position.
    BACKWARD_POLICY: A policy that represents moving backward, with a value of -1 in the first position.
    LEFT_POLICY: A policy that represents moving left, with a value of -1 in the second position.
    RIGHT_POLICY: A policy that represents moving right, with a value of 1 in the second position.
    ANTI_CLOCKWISE_POLICY: A policy that represents turning anti-clockwise, with a value of -1 in the third position.
    CLOCKWISE_POLICY: A policy that represents turning clockwise, with a value of 1 in the third position.

Worlds:
    EMPTY_WORLD: An empty world with a single agent.
    TWO_AGENT_EMPTY_WORLD: An empty world with two agents.
    TWO_AGENT_EMPTY_WORLD_COLLIDE: An empty world with two agents that collide.
    RAMP_30: A world with a 30 degree ramp.
    RAMP_45: A world with a 45 degree ramp.
    RAMP_60: A world with a 60 degree ramp.

Frames:
    TEST_TILEMAP_FLAT: A 10x10 flat world of height 3.
    TEST_TILEMAP_BUMP: A 10x10 flat world but with a bump of height 5 at z=5.
    TEST_TILEMAP_WORLD42: A pre-generated world using seed 42 with ramps and terrain.
    TEST_FRAME: A 10x10 low-res frame of the rendered tilemap.
    TEST_FRAME_WITH_PROPS: A 10x10 low-res frame of the render with props.
"""

import jax.numpy as jnp

# ========================
# Policy Definitions
# ========================

NOOP_POLICY = jnp.array([0, 0, 0, 0, 0, 0])
FORWARD_POLICY = jnp.array([1, 0, 0, 0, 0, 0])
BACKWARD_POLICY = jnp.array([-1, 0, 0, 0, 0, 0])
LEFT_POLICY = jnp.array([0, 1, 0, 0, 0, 0])
RIGHT_POLICY = jnp.array([0, -1, 0, 0, 0, 0])
ANTI_CLOCKWISE_POLICY = jnp.array([0, 0, -1, 0, 0, 0])
CLOCKWISE_POLICY = jnp.array([0, 0, 1, 0, 0, 0])
LOOK_UP_POLICY = jnp.array([0, 0, 0, 1, 0, 0])
LOOK_DOWN_POLICY = jnp.array([0, 0, 0, -1, 0, 0])
TAG_AGENT_POLICY = jnp.array([0, 0, 0, 0, 1, 0])
GRAB_AGENT_POLICY = jnp.array([0, 0, 0, 0, 0, 1])

# ========================
# World Definitions
# ========================

EMPTY_WORLD = """
<mujoco>
  <worldbody>
    <light name="main_light" pos="0 0 1" dir="0 0 -1"
           diffuse="1 1 1" specular="0.1 0.1 0.1"/>

    <body name="agent0" pos="0 0 1">
      <joint type="slide" axis="1 0 0" />
      <joint type="slide" axis="0 1 0" />
      <joint type="slide" axis="0 0 1" />
      <joint type="hinge" axis="0 0 1" stiffness="1"/>

      <geom
        name="agent0_geom"
        type="capsule"
        size="0.06"
        fromto="0 0 0 0 0 -0.4"
        
        mass="1"
      />

      <geom
        type="box"
        size="0.05 0.05 0.05"
        pos="0 0 0.2"
        
        mass="0"
      />
    </body>

    <!-- Ground plane, also with low friction -->
    <geom
      name="ground"
      type="plane"
      size="0 0 0.01"
      rgba="1 1 1 1"
    />

  </worldbody>
</mujoco>
"""

TWO_AGENT_EMPTY_WORLD = """
<mujoco>
  <worldbody>
    <light name="main_light" pos="0 0 1" dir="0 0 -1"
           diffuse="1 1 1" specular="0.1 0.1 0.1"/>

    <body name="agent0" pos="0 0 1">
      <joint type="slide" axis="1 0 0" />
      <joint type="slide" axis="0 1 0" />
      <joint type="slide" axis="0 0 1" />
      <joint type="hinge" axis="0 0 1" stiffness="1"/>

      <geom
        name="agent0_geom"
        type="capsule"
        size="0.06"
        fromto="0 0 0 0 0 -0.4"
        
        mass="1"
      />

      <geom
        type="box"
        size="0.05 0.05 0.05"
        pos="0 0 0.2"
        
        mass="0"
      />
    </body>

    <body name="agent1" pos="0 0.5 1">
      <joint type="slide" axis="1 0 0" />
      <joint type="slide" axis="0 1 0" />
      <joint type="slide" axis="0 0 1" />
      <joint type="hinge" axis="0 0 1" stiffness="1"/>

      <geom
        name="agent1_geom"
        type="capsule"
        size="0.06"
        fromto="0 0 0 0 0 -0.4"
        
        mass="1"
      />

      <geom
        type="box"
        size="0.05 0.05 0.05"
        pos="0 0 0.2"
        
        mass="0"
      />
    </body>

    <!-- Ground plane, also with low friction -->
    <geom
      name="ground"
      type="plane"
      size="0 0 0.01"
      rgba="1 1 1 1"
    />

  </worldbody>
</mujoco>
"""

TWO_AGENT_EMPTY_WORLD_COLLIDE = """
<mujoco>
  <worldbody>
    <light name="main_light" pos="0 0 1" dir="0 0 -1"
           diffuse="1 1 1" specular="0.1 0.1 0.1"/>

    <body name="agent0" pos="0 0 1">
      <joint type="slide" axis="1 0 0" />
      <joint type="slide" axis="0 1 0" />
      <joint type="slide" axis="0 0 1" />
      <joint type="hinge" axis="0 0 1" stiffness="1"/>

      <geom
        name="agent0_geom"
        type="capsule"
        size="0.06"
        fromto="0 0 0 0 0 -0.4"
        mass="1"
      />

      <geom
        type="box"
        size="0.05 0.05 0.05"
        pos="0 0 0.2"
        mass="0"
      />
    </body>

    <body name="agent1" pos="1 0 1">
      <joint type="slide" axis="1 0 0" />
      <joint type="slide" axis="0 1 0" />
      <joint type="slide" axis="0 0 1" />
      <joint type="hinge" axis="0 0 1" stiffness="1"/>

      <geom
        name="agent1_geom"
        type="capsule"
        size="0.06"
        fromto="0 0 0 0 0 -0.4"
        mass="1"
      />

      <geom
        type="box"
        size="0.05 0.05 0.05"
        pos="0 0 0.2"
        mass="0"
      />
    </body>

    <!-- Ground plane, also with low friction -->
    <geom
      name="ground"
      type="plane"
      size="0 0 0.01"
      rgba="1 1 1 1"
    />

  </worldbody>
</mujoco>
"""

RAMP_30 = """
<mujoco>
  <worldbody>
    <light name="main_light" pos="0 0 1" dir="0 0 -1"
           diffuse="1 1 1" specular="0.1 0.1 0.1"/>

    <body name="agent0" pos="0 0 1">
      <joint type="slide" axis="1 0 0" />
      <joint type="slide" axis="0 1 0" />
      <joint type="slide" axis="0 0 1" />
      <joint type="hinge" axis="0 0 1" />

      <geom
        name="agent0_geom"
        type="capsule"
        size="0.06"
        fromto="0 0 0 0 0 -0.4"
        mass="1"
      />

      <geom
        type="box"
        size="0.05 0.05 0.05"
        pos="0 0 0.2"
        mass="0"
      />
    </body>

    <!-- Ground plane, also with low friction -->
    <geom
      name="ground"
      type="plane"
      size="0 0 0.01"
      rgba="1 1 1 1"
    />

    <geom type="box" size="1 1 1" pos="2 0 -0.5" euler="0 60 0" rgba="1 0.8 0.8 1" />

  </worldbody>
</mujoco>
"""

RAMP_45 = """
<mujoco>
  <worldbody>
    <light name="main_light" pos="0 0 1" dir="0 0 -1"
           diffuse="1 1 1" specular="0.1 0.1 0.1"/>

    <body name="agent0" pos="0 0 1">
      <joint type="slide" axis="1 0 0" />
      <joint type="slide" axis="0 1 0" />
      <joint type="slide" axis="0 0 1" />
      <joint type="hinge" axis="0 0 1" stiffness="1"/>

      <geom
        name="agent0_geom"
        type="capsule"
        size="0.06"
        fromto="0 0 0 0 0 -0.4"
        
        mass="1"
      />

      <geom
        type="box"
        size="0.05 0.05 0.05"
        pos="0 0 0.2"
        
        mass="0"
      />
    </body>

    <!-- Ground plane, also with low friction -->
    <geom
      name="ground"
      type="plane"
      size="0 0 0.01"
      rgba="1 1 1 1"
    />

    <geom type="box" size="1 1 1" pos="2 0 -0.5" euler="0 45 0" rgba="1 0.8 0.8 1" />

  </worldbody>
</mujoco>
"""

RAMP_60 = """
<mujoco>
  <worldbody>
    <light name="main_light" pos="0 0 1" dir="0 0 -1"
           diffuse="1 1 1" specular="0.1 0.1 0.1"/>

    <body name="agent0" pos="0 0 1">
      <joint type="slide" axis="1 0 0" />
      <joint type="slide" axis="0 1 0" />
      <joint type="slide" axis="0 0 1" />
      <joint type="hinge" axis="0 0 1" stiffness="1"/>

      <geom
        name="agent0_geom"
        type="capsule"
        size="0.06"
        fromto="0 0 0 0 0 -0.4"
        
        mass="1"
      />
    </body>

    <!-- Ground plane, also with low friction -->
    <geom
      name="ground"
      type="plane"
      size="0 0 0.01"
      rgba="1 1 1 1"
    />

    <geom type="box" size="1 1 1" pos="2 0 -0.5" euler="0 30 0" rgba="1 0.8 0.8 1" name="ramp" />

  </worldbody>
</mujoco>
"""
DEFAULT_VIEWSIZE: tuple[int, int] = (96, 72)

TEST_TILEMAP_EMPTY_WORLD = jnp.array(
    [[[0, 0, 0, 0] for _ in range(10)] for _ in range(10)]
)
TEST_TILEMAP_FLAT = jnp.array([[[0, 0, 0, 3] for _ in range(10)] for _ in range(10)])
TEST_TILEMAP_BUMP = TEST_TILEMAP_FLAT.at[:, 5].set(jnp.array([0, 0, 0, 5]))
TEST_TILEMAP_WORLD42 = jnp.array(
    [
        [
            [0, 0, 0, 2],
            [0, 2, 0, 2],
            [0, 2, 0, 2],
            [0, 1, 0, 2],
            [0, 0, 0, 5],
            [0, 2, 0, 5],
            [0, 1, 0, 5],
            [0, 0, 0, 3],
            [0, 2, 0, 3],
            [0, 1, 0, 3],
        ],
        [
            [0, 3, 0, 3],
            [0, 0, 0, 3],
            [0, 2, 0, 3],
            [0, 3, 0, 4],
            [0, 2, 0, 4],
            [0, 3, 0, 5],
            [0, 0, 0, 5],
            [0, 2, 0, 5],
            [0, 3, 0, 6],
            [0, 2, 0, 6],
        ],
        [
            [0, 1, 0, 3],
            [0, 0, 0, 3],
            [1, 1, 3, 4],
            [0, 0, 0, 4],
            [0, 1, 0, 4],
            [0, 0, 0, 5],
            [0, 2, 0, 5],
            [0, 1, 0, 5],
            [0, 0, 0, 6],
            [0, 1, 0, 6],
        ],
        [
            [1, 3, 3, 4],
            [0, 0, 0, 3],
            [0, 3, 0, 3],
            [0, 3, 0, 6],
            [0, 2, 0, 6],
            [0, 3, 0, 4],
            [0, 0, 0, 4],
            [1, 2, 4, 5],
            [0, 0, 0, 4],
            [0, 2, 0, 4],
        ],
        [
            [0, 1, 0, 3],
            [0, 0, 0, 3],
            [0, 3, 0, 3],
            [0, 0, 0, 6],
            [0, 1, 0, 6],
            [0, 1, 0, 4],
            [0, 0, 0, 4],
            [0, 0, 0, 4],
            [0, 0, 0, 4],
            [0, 3, 0, 4],
        ],
        [
            [1, 3, 3, 4],
            [0, 0, 0, 3],
            [0, 0, 0, 3],
            [0, 2, 0, 3],
            [0, 3, 0, 4],
            [0, 1, 0, 4],
            [0, 0, 0, 4],
            [0, 3, 0, 4],
            [0, 2, 0, 4],
            [0, 1, 0, 4],
        ],
        [
            [0, 1, 0, 3],
            [0, 0, 0, 3],
            [0, 0, 0, 3],
            [0, 3, 0, 3],
            [0, 1, 0, 4],
            [0, 0, 0, 4],
            [0, 3, 0, 4],
            [2, 1, 4, 6],
            [0, 0, 0, 6],
            [0, 2, 0, 6],
        ],
        [
            [0, 1, 0, 3],
            [0, 0, 0, 3],
            [0, 0, 0, 3],
            [1, 1, 3, 4],
            [0, 1, 0, 4],
            [0, 0, 0, 4],
            [0, 3, 0, 4],
            [0, 1, 0, 6],
            [0, 0, 0, 6],
            [0, 3, 0, 6],
        ],
        [
            [0, 0, 0, 3],
            [0, 2, 0, 3],
            [0, 2, 0, 3],
            [0, 1, 0, 3],
            [0, 0, 0, 4],
            [0, 2, 0, 4],
            [0, 1, 0, 4],
            [0, 0, 0, 6],
            [0, 2, 0, 6],
            [0, 1, 0, 6],
        ],
        [
            [0, 3, 0, 2],
            [0, 0, 0, 2],
            [0, 0, 0, 2],
            [0, 2, 0, 2],
            [0, 3, 0, 5],
            [0, 0, 0, 5],
            [0, 2, 0, 5],
            [0, 3, 0, 3],
            [0, 0, 0, 3],
            [0, 2, 0, 3],
        ],
    ]
)
TEST_TILEMAP_MAX_HEIGHT = jnp.array(
    [[[0, 0, 0, 6] for _ in range(10)] for _ in range(10)]
)
TEST_FRAME = jnp.array(
    [
        [
            [0.701171875, 0.701171875, 0.701171875],
            [0.63037109375, 0.63037109375, 0.63037109375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.70458984375, 0.70458984375, 0.70458984375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.70458984375, 0.70458984375, 0.70458984375],
        ],
        [
            [0.26025390625, 0.360107421875, 0.47216796875],
            [0.26025390625, 0.3603515625, 0.472412109375],
            [0.26025390625, 0.3603515625, 0.472412109375],
            [0.26025390625, 0.3603515625, 0.472412109375],
            [0.260009765625, 0.360107421875, 0.47216796875],
            [0.70458984375, 0.70458984375, 0.70458984375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.61474609375, 0.61474609375, 0.61474609375],
            [0.70458984375, 0.70458984375, 0.70458984375],
        ],
        [
            [0.268798828125, 0.372314453125, 0.488037109375],
            [0.265625, 0.367919921875, 0.482177734375],
            [0.2626953125, 0.36376953125, 0.47705078125],
            [0.26025390625, 0.3603515625, 0.472412109375],
            [0.2587890625, 0.3583984375, 0.4697265625],
            [0.5859375, 0.5859375, 0.5859375],
            [0.5927734375, 0.5927734375, 0.5927734375],
            [0.70068359375, 0.70068359375, 0.70068359375],
            [0.703125, 0.703125, 0.703125],
            [0.58642578125, 0.58642578125, 0.58642578125],
        ],
        [
            [0.260009765625, 0.360107421875, 0.47216796875],
            [0.26025390625, 0.3603515625, 0.472412109375],
            [0.26025390625, 0.3603515625, 0.472412109375],
            [0.259765625, 0.359619140625, 0.4716796875],
            [0.26025390625, 0.3603515625, 0.472412109375],
            [0.38134765625, 0.38134765625, 0.38134765625],
            [0.70458984375, 0.70458984375, 0.70458984375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.70458984375, 0.70458984375, 0.70458984375],
        ],
        [
            [0.26025390625, 0.3603515625, 0.472412109375],
            [0.260009765625, 0.35986328125, 0.471923828125],
            [0.26025390625, 0.3603515625, 0.472412109375],
            [0.26025390625, 0.360107421875, 0.47216796875],
            [0.26025390625, 0.3603515625, 0.472412109375],
            [0.38134765625, 0.38134765625, 0.38134765625],
            [0.62158203125, 0.62158203125, 0.62158203125],
            [0.70458984375, 0.70458984375, 0.70458984375],
            [0.69677734375, 0.69677734375, 0.69677734375],
            [0.5859375, 0.5859375, 0.5859375],
        ],
        [
            [0.317138671875, 0.317138671875, 0.317138671875],
            [0.38134765625, 0.38134765625, 0.38134765625],
            [0.317138671875, 0.317138671875, 0.317138671875],
            [0.319580078125, 0.319580078125, 0.319580078125],
            [0.38134765625, 0.38134765625, 0.38134765625],
            [0.317138671875, 0.317138671875, 0.317138671875],
            [0.66748046875, 0.66748046875, 0.66748046875],
            [0.70458984375, 0.70458984375, 0.70458984375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.70458984375, 0.70458984375, 0.70458984375],
        ],
        [
            [0.70703125, 0.70703125, 0.70703125],
            [0.587890625, 0.587890625, 0.587890625],
            [0.705078125, 0.705078125, 0.705078125],
            [0.5859375, 0.5859375, 0.5859375],
            [0.70458984375, 0.70458984375, 0.70458984375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.70458984375, 0.70458984375, 0.70458984375],
            [0.5859375, 0.5859375, 0.5859375],
            [0.70458984375, 0.70458984375, 0.70458984375],
            [0.5859375, 0.5859375, 0.5859375],
        ],
        [
            [0.6689453125, 0.6689453125, 0.6689453125],
            [0.6650390625, 0.6650390625, 0.6650390625],
            [0.65478515625, 0.65478515625, 0.65478515625],
            [0.6494140625, 0.6494140625, 0.6494140625],
            [0.6484375, 0.6484375, 0.6484375],
            [0.6484375, 0.6484375, 0.6484375],
            [0.6484375, 0.6484375, 0.6484375],
            [0.6484375, 0.6484375, 0.6484375],
            [0.6484375, 0.6484375, 0.6484375],
            [0.6484375, 0.6484375, 0.6484375],
        ],
        [
            [0.7724609375, 0.7724609375, 0.7724609375],
            [0.6640625, 0.6640625, 0.6640625],
            [0.7314453125, 0.7314453125, 0.7314453125],
            [0.70849609375, 0.70849609375, 0.70849609375],
            [0.58837890625, 0.58837890625, 0.58837890625],
            [0.703125, 0.703125, 0.703125],
            [0.587890625, 0.587890625, 0.587890625],
            [0.587890625, 0.587890625, 0.587890625],
            [0.703125, 0.703125, 0.703125],
            [0.587890625, 0.587890625, 0.587890625],
        ],
        [
            [0.8427734375, 0.8427734375, 0.8427734375],
            [0.740234375, 0.740234375, 0.740234375],
            [0.76953125, 0.76953125, 0.76953125],
            [0.7080078125, 0.7080078125, 0.7080078125],
            [0.587890625, 0.587890625, 0.587890625],
            [0.70458984375, 0.70458984375, 0.70458984375],
            [0.60107421875, 0.60107421875, 0.60107421875],
            [0.5859375, 0.5859375, 0.5859375],
            [0.70458984375, 0.70458984375, 0.70458984375],
            [0.58642578125, 0.58642578125, 0.58642578125],
        ],
    ]
)
TEST_FRAME_WITH_PROPS = jnp.array(
    [
        [
            [0.7046296000480652, 0.7046296000480652, 0.7046296000480652],
            [0.5860296487808228, 0.5860296487808228, 0.5860296487808228],
            [0.6928065419197083, 0.6928065419197083, 0.6928065419197083],
            [0.5860294103622437, 0.5860294103622437, 0.5860294103622437],
            [0.7046296000480652, 0.7046296000480652, 0.7046296000480652],
            [0.7046296000480652, 0.7046296000480652, 0.7046296000480652],
            [0.5860294103622437, 0.5860294103622437, 0.5860294103622437],
            [0.5871762037277222, 0.5871762037277222, 0.5871762037277222],
            [0.7039504051208496, 0.7039504051208496, 0.7039504051208496],
            [0.6969494819641113, 0.6969494819641113, 0.6969494819641113],
        ],
        [
            [0.0, 0.4805370569229126, 0.0],
            [0.0, 0.48099929094314575, 0.0],
            [0.0, 0.48104530572891235, 0.0],
            [0.0, 0.4810141324996948, 0.0],
            [0.0, 0.4808400869369507, 0.0],
            [0.5860294103622437, 0.5860294103622437, 0.5860294103622437],
            [0.704594612121582, 0.704594612121582, 0.704594612121582],
            [0.7046043872833252, 0.7046043872833252, 0.7046043872833252],
            [0.7045301795005798, 0.7045301795005798, 0.7045301795005798],
            [0.5865076184272766, 0.5865076184272766, 0.5865076184272766],
        ],
        [
            [3.271687819506042e-05, 0.8887746334075928, 3.271687819506042e-05],
            [8.903251000447199e-06, 0.8886584043502808, 8.903251000447199e-06],
            [1.3764869208898745e-06, 0.8887394666671753, 1.3764869208898745e-06],
            [1.2042305286286137e-07, 0.8887507915496826, 1.2042305286286137e-07],
            [6.399829999281792e-09, 0.8887008428573608, 6.399829999281792e-09],
            [0.5860294103622437, 0.5860294103622437, 0.5860294103622437],
            [0.5860294103622437, 0.5860294103622437, 0.5860294103622437],
            [0.5862460136413574, 0.5862460136413574, 0.5862460136413574],
            [0.7046296000480652, 0.7046296000480652, 0.7046296000480652],
            [0.7046296000480652, 0.7046296000480652, 0.7046296000480652],
        ],
        [
            [0.0, 0.47988688945770264, 0.0],
            [0.0, 0.4810419976711273, 0.0],
            [0.0, 0.4810386896133423, 0.0],
            [0.0, 0.47697511315345764, 0.0],
            [0.0, 0.4810387194156647, 0.0],
            [0.38145747780799866, 0.38145747780799866, 0.38145747780799866],
            [0.7046296000480652, 0.7046296000480652, 0.7046296000480652],
            [0.5860320925712585, 0.5860320925712585, 0.5860320925712585],
            [0.5860294103622437, 0.5860294103622437, 0.5860294103622437],
            [0.5866138339042664, 0.5866138339042664, 0.5866138339042664],
        ],
        [
            [0.0, 0.4810403883457184, 0.0],
            [0.0, 0.48057714104652405, 0.0],
            [0.0, 0.48106497526168823, 0.0],
            [0.0, 0.4808911979198456, 0.0],
            [0.0, 0.4810304641723633, 0.0],
            [0.3814574182033539, 0.3814574182033539, 0.3814574182033539],
            [0.6268107891082764, 0.6268107891082764, 0.6268107891082764],
            [0.7046295404434204, 0.7046295404434204, 0.7046295404434204],
            [0.5860300660133362, 0.5860300660133362, 0.5860300660133362],
            [0.5860294699668884, 0.5860294699668884, 0.5860294699668884],
        ],
        [
            [0.31725218892097473, 0.31725218892097473, 0.31725218892097473],
            [0.3814574182033539, 0.3814574182033539, 0.3814574182033539],
            [0.31725215911865234, 0.31725215911865234, 0.31725215911865234],
            [0.31727030873298645, 0.31727030873298645, 0.31727030873298645],
            [0.3814573884010315, 0.3814573884010315, 0.3814573884010315],
            [0.31725215911865234, 0.31725215911865234, 0.31725215911865234],
            [0.6534721255302429, 0.6534721255302429, 0.6534721255302429],
            [0.7046296000480652, 0.7046296000480652, 0.7046296000480652],
            [0.5860294103622437, 0.5860294103622437, 0.5860294103622437],
            [0.7046296000480652, 0.7046296000480652, 0.7046296000480652],
        ],
        [
            [0.7068762183189392, 0.7068762183189392, 0.7068762183189392],
            [0.5878639817237854, 0.5878639817237854, 0.5878639817237854],
            [0.7050284147262573, 0.7050284147262573, 0.7050284147262573],
            [0.5860771536827087, 0.5860771536827087, 0.5860771536827087],
            [0.7046307921409607, 0.7046307921409607, 0.7046307921409607],
            [0.5860294103622437, 0.5860294103622437, 0.5860294103622437],
            [0.7046296000480652, 0.7046296000480652, 0.7046296000480652],
            [0.5860294103622437, 0.5860294103622437, 0.5860294103622437],
            [0.7046296000480652, 0.7046296000480652, 0.7046296000480652],
            [0.5860294103622437, 0.5860294103622437, 0.5860294103622437],
        ],
        [
            [0.7228245735168457, 0.7228245735168457, 0.7228245735168457],
            [0.7195366621017456, 0.7195366621017456, 0.7195366621017456],
            [0.5928553342819214, 0.5928553342819214, 0.5928553342819214],
            [0.7054046392440796, 0.7054046392440796, 0.7054046392440796],
            [0.5860785841941833, 0.5860785841941833, 0.5860785841941833],
            [0.7046298384666443, 0.7046298384666443, 0.7046298384666443],
            [0.5860299468040466, 0.5860299468040466, 0.5860299468040466],
            [0.704629123210907, 0.704629123210907, 0.704629123210907],
            [0.5860580205917358, 0.5860580205917358, 0.5860580205917358],
            [0.5861576795578003, 0.5861576795578003, 0.5861576795578003],
        ],
        [
            [0.6698926091194153, 0.6698926091194153, 0.6698926091194153],
            [0.7672428488731384, 0.7672428488731384, 0.7672428488731384],
            [0.620652973651886, 0.620652973651886, 0.620652973651886],
            [0.5925986766815186, 0.5925986766815186, 0.5925986766815186],
            [0.7050103545188904, 0.7050103545188904, 0.7050103545188904],
            [0.586046040058136, 0.586046040058136, 0.586046040058136],
            [0.7046273350715637, 0.7046273350715637, 0.7046273350715637],
            [0.7046270370483398, 0.7046270370483398, 0.7046270370483398],
            [0.5860323905944824, 0.5860323905944824, 0.5860323905944824],
            [0.7046271562576294, 0.7046271562576294, 0.7046271562576294],
        ],
        [
            [0.7689034342765808, 0.7689034342765808, 0.7689034342765808],
            [0.7401685118675232, 0.7401685118675232, 0.7401685118675232],
            [0.7697005867958069, 0.7697005867958069, 0.7697005867958069],
            [0.651146650314331, 0.651146650314331, 0.651146650314331],
            [0.5878591537475586, 0.5878591537475586, 0.5878591537475586],
            [0.7046911716461182, 0.7046911716461182, 0.7046911716461182],
            [0.662687361240387, 0.662687361240387, 0.662687361240387],
            [0.5860294103622437, 0.5860294103622437, 0.5860294103622437],
            [0.7046296000480652, 0.7046296000480652, 0.7046296000480652],
            [0.6845281720161438, 0.6845281720161438, 0.6845281720161438],
        ],
    ]
)
